// Things that could cause glitching:
// - A broken address line somewhere or other hardware bug
// - Incompatbility between um6502 and mos6510
// - Damaged SID.
// - Something unexpected in the memory map of the player -- specifically
//   check the zero page.
// - The fact that the more complicated the tune gets the more out of sync
//   it gets, duh (the problem with a hardcoded delay).
//   - Going to need a real periodic interrupt source (or a RTC to busy-
//     wait against)
//   - Could just hack it by raising the IRQ line for one cycle only, 
//     easy enough from asm, if we actually had a timer
//     - just use a spare PIC line -- perhaps repurpose OSCOUT.
//     - Perhaps set up a timer as Joel suggests -- highest possible frequency
//       divider, and count inside.
//     - Slightly dodgier alternative is to set up a timer (timer0?) as a counter,
//       and check it while busy-waiting for input.
//     - Specifically, timer0, prescaler 4, 16-bit mode, counter set at
//       (65536 - 16666), gives an interrupt 1/60th of a second in the future.
//     - The same with a counter at (65536 - 20000) gives an interrupt 1/50th
//       of a second in the future. The IRQ routine can raise the VBL interrupt,
//       wait 16 cycles, lower the VBL interrupt, and re-set the counter.

#include <stdio.h>
#include <usart.h>
#include "pic18fregs.h"

// non-free/include/pic16/pic18f14k50.h
// Magic words courtesy http://comments.gmane.org/gmane.comp.compilers.sdcc.user/887
// (then hacked up by wzdd)
__CONFIG(__CONFIG1L, _CPUDIV_NOCLKDIV_1L);
// Produce a phi / 4 clock on CLKOUT. Currently not used (we set up 16Mhz below,
// so this gives 4 MHz)
__CONFIG(__CONFIG1H, _FOSC_IRCCLKOUT_1H);
__CONFIG(__CONFIG2L, _BOREN_OFF_2L);
__CONFIG(__CONFIG2H, _WDTEN_OFF_2H);
// Master Clear!
__CONFIG(__CONFIG3H, _MCLRE_OFF_3H);
// LVP_OFF_4L is important.
__CONFIG(__CONFIG4L, _XINST_OFF_4L & _LVP_OFF_4L);

__CONFIG(__CONFIG5L, _CP0_OFF_5L & _CP1_OFF_5L);
__CONFIG(__CONFIG5H, _CPD_OFF_5H & _CPB_OFF_5H);
__CONFIG(__CONFIG6L, _WRT0_OFF_6L & _WRT1_OFF_6L);
__CONFIG(__CONFIG6H, _WRTD_OFF_6H & _WRTB_OFF_6H & _WRTC_OFF_6H); 
__CONFIG(__CONFIG7L, _EBTR0_OFF_7L & _EBTR1_OFF_7L);
__CONFIG(__CONFIG7H, _EBTRB_OFF_7H);

// Delay about 500 nanoseconds.
#define WAITHALF __asm \
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	__endasm;

// Delay about 250 nanoseconds.
#define WAITQUARTER __asm \
	nop\n\
	nop\n\
	nop\n\
	nop\n\
	__endasm;

#define NOP __asm \
	nop\n \
	__endasm;

// Overflow after (65536-20000) timer ticks = 50 Hz
#define TMR1H_50HZ 0xb1
#define TMR1L_50HZ 0xe0
// Overflow after (65536-16666) timer ticks = 60 Hz
#define TMR1H_60HZ 0xbe
#define TMR1L_60HZ 0xe6

volatile unsigned char tmr1hval, tmr1lval;
volatile unsigned char tmr1rd16, tmr1tickps1, tmr1tickps0, tmr1cs, tmr1en;

#define BUFSIZE 256
unsigned char buffer[BUFSIZE];

#define HASHLENGTH 40
unsigned char hash[HASHLENGTH];

// If enabled, all reads go to the ROM and writes go to RAM or SID.
// If disabled, all reads and writes go to RAM or SID.
#define ROMENABLE  LATBbits.LATB6

// Controls the 4094 8-bit latch which connects us to the data bus.
#define DATADATA   LATCbits.LATC6
#define DATASTROBE LATCbits.LATC4
#define DATACLOCK  LATCbits.LATC3

// Controls the 6502 clock. This is toggled "by hand" when loading from
// ROM, and connected to PWM1 when running the CPU at full speed.
#define CLOCK6502  LATCbits.LATC5

// Connected to 6502 /RES. Pull low to reset the processor.
#define RESET6502  LATCbits.LATC2

// A configurable interrupt to drive music playing
#define INT6502    LATCbits.LATC7

void puts(char *s)
{
	char *c = s;
	while(*c) {
		while(TXSTAbits.TRMT == 0)
			;
		TXREG = *c;
		c++;
	}
}

unsigned char getc(void)
{
	unsigned char c;

	TXSTAbits.SYNC = 0;
	RCSTAbits.SPEN = 1;
	RCSTAbits.RX9  = 0;

	// Clear framing errors.
	if(RCSTAbits.FERR) {
		// Reset UART.
		RCSTAbits.SPEN = 0;
		RCSTAbits.SPEN = 1;
	}


	// Wait for a character to enter the FIFO.
	while(PIR1bits.RCIF == 0)
		;

	// Read the character (this clears it from the FIFO).
	c = RCREG;

	// Handle any buffer overrun errors.
	if(RCSTAbits.OERR) {
		// Clear FIFO
		RCSTAbits.CREN = 0;
		RCSTAbits.CREN = 1;
	}

	return c;
}

/* This isn't used any more. See comment below for why! */
void pwm_start(int hz)
{
	// From http://www.micro-examples.com/public/microex-navig/doc/097-pwm-calculator.html
	// At 16MHz, we can't generate 50 or 60 hz (which is what we want) directly.
	// Instead we generate it multiplied by 32, and then clock-divide the result
	// in hardware.
	//
	// ... and, aargh, 6502's /IRQ is level sensitive! Means we can't do this,
	// basically. Have to use a busy loop in the driver routine.
	switch(hz) {
	case 50:
		PR2 = 0b10011011 ;
		T2CON = 0b00000111 ;
		CCPR1L = 0b01001101 ;
		CCP1CON = 0b00111100 ;
		break;
	default: // 60
		PR2 = 0b10000001 ;
		T2CON = 0b00000111 ;
		CCPR1L = 0b01000000 ;
		CCP1CON = 0b00111100 ;
		break;
	}
}

void data_bus_write(unsigned char b)
{
	int i;

	// Note the lack of explicit timing here. 4094s aren't very fast.
	// But it seems to work at 16MHz.
	DATASTROBE = 0; // Strobe low, store new data
	NOP
	NOP
	DATASTROBE = 1; // Strobe high, put data on output pins.

	for(i = 0; i < 8; i++) {
		DATACLOCK = 0; // Clock low
		if((b & 128) != 0)
			DATADATA = 1;
		else 
			DATADATA = 0;
		DATACLOCK = 1; // Clock high
		b <<= 1;
	}

	DATASTROBE = 0; // Strobe low, store new data
	NOP
	NOP
	DATASTROBE = 1; // Strobe high, put data on output pins.

}

void cpu_clock_full(int cycles)
{
	//unsigned char addr;
	while(cycles){
		cycles --;
		//addr = PORTC & 0x3;
		//printf("clock full: pre: %x\r\n", addr);
		NOP
		NOP
		NOP
		NOP
		// Clock low.
		CLOCK6502 = 0;

		NOP
		NOP
		NOP
		NOP
		//addr = PORTC & 0x3;
		//printf("clock full: mid: %x\r\n", addr);

		// Clock high.
		CLOCK6502 = 1;
	}

}

// Run the CPU clock at full speed (a blazing 1MHz) by enabling the 
// PIC's PWM module. At 16MHz, this produces a 1MHz clock with 50%
// duty cycle. (See http://www.micro-examples.com/public/microex-navig/doc/097-pwm-calculator.html)
void cpu_clock_run(int shouldRun)
{
	if(shouldRun) {
		// Turn on PWM, 1Mhz.
		PR2 = 0b00000000 ;
		T2CON = 0b00000101 ;
		CCPR1L = 0b00000000 ;
		CCP1CON = 0b00011100 ;
	} else {
		// Turn off PWM:
		CCP1CON = 0b00000000 ;
	}
}

// Start the CPU and get it ready to read program code at 0xFF00.
int cpu_boot()
{
	unsigned char addr;
	// Hold reset.
	RESET6502 = 0;
	cpu_clock_full(2);

	//printf("Bring reset up\r\n");
	// Bring reset up and wait six cycles for the CPU to start.
	RESET6502 = 1;
	cpu_clock_full(6);

	addr = PORTC & 0x3;
	//printf("Pre half cycle %x\r\n", addr);
	// Clock the address lines in -- first half of clock
	CLOCK6502 = 0;
	WAITHALF

	// Read our address lines.
	addr = PORTC & 0x3;
	if(addr != 0) {
		printf("First address read fail %x\r\n", addr);
		RESET6502 = 0;
		return -1;
	}

	// The CPU is trying to read 0xFFFC, which is the lower 8 bits of
	// the reset address.
	data_bus_write(0);

	// Second half of 6502 clock cycle.
	CLOCK6502 = 1;
	WAITHALF

	// Clock the CPU again.
	CLOCK6502 = 0;
	WAITHALF;

	addr = PORTC & 0x3;
	if(addr != 1) {
		printf("Second address read fail %x\r\n", addr);
		RESET6502 = 0;
		return -1;
	}

	// The CPU is now trying to read 0xFFFD, which is the upper 8 bits of
	// the reset address.
	data_bus_write(0xf0);
	CLOCK6502 = 1;
	WAITHALF;

	return 0;
}

void cpu_halt()
{
	RESET6502 = 0;
	cpu_clock_full(2);
}

void cpu_test()
{
	if(cpu_boot() != 0)
		return;

	cpu_halt();
	
	printf("CPU booted OK\r\n");
}

unsigned char sidProgram[] = {
	/* Adapted from a SID test routine by Michael Kohn. */
	/* Set volume (3, max 15) */
	0xa9, 3,			/* lda 3 */  /* 10101001 00000011 */ /* 1001, 0011 */
	0x8d, 0x18, 0xd4,	/* sta 0xd418 */
	0xd4,				/* 1 wait state */
	/* Set attack / decay, sustain / release */
	0xa9, 0xbe,			/* lda 0xbe */
	0x8d, 0x05, 0xd4,	/* sta 0xd405 */
	0xd4,				/* 1 wait state */
	0xa9, 0xf8,			/* lda 0xf8 */
	0x8d, 0x06, 0xd4,	/* sta 0xd406 */
	0xd4,				/* 1 wait state */
	/* Set frequency to A3 */
	0xa9, 0x18,			/* lda 0x18 */
	0x8d, 0x00, 0xd4,	/* sta 0xd400 */
	0xd4,				/* 1 wait state */
	0xa9, 0xe,			/* lda 0xe */
	0x8d, 0x01, 0xd4,	/* sta 0xd401 */
	0xd4,				/* 1 wait state */
	/* Square waveform on voice 1 */
	0xa9, 0x21,			/* lda 0x11 */
	0x8d, 0x04, 0xd4,	/* sta 0xd404 */
	0xd4,				/* 1 wait state */
};
void sid_test()
{
	unsigned int i;
	unsigned long j;
	
	if(cpu_boot() != 0)
		return ;

	// Now the CPU should start requesting data from 0xf000. We'll see this
	// as increasing requests on the address lines. 
	
	for(i = 0; i < 36; i++) {
		unsigned char codeByte = sidProgram[i];
		CLOCK6502 = 0;
		WAITHALF;
		// We always put stuff onto "the data bus", which is OK because
		// we're actually writing to a 4094, which will be tristated by
		// the address decoder / CPU if it's not supposed to be driving
		// the bus. FIXME: Check OE on the 4094.
		data_bus_write(codeByte);
		CLOCK6502 = 1;
		WAITHALF;
	}

	// Run the CPU at full speed for a little while.
	cpu_clock_run(1);
	for(j = 0; j < 100000; j++)
		;
	cpu_clock_run(0);

	// Put the CPU and SID back into reset state.
	cpu_halt();

	printf("SID test done. You should have heard a horrible noise.\r\n");
}

void bus_test(int num)
{
	unsigned char val = 0;
	switch(num) {
		case '0':
			val = 0x0;
			break;
		case '1':
			val = 0xff;
			break;
		case '2':
			val = 0xAA;
			break;
	}

	ROMENABLE = 1;
	data_bus_write(val);
}

void upload()
{
	int startAddress = 0, length = 0;
	int i;
	//unsigned char addr;
	unsigned short destAddrHigh;
	
	// Start address: two bytes.
	startAddress = (getc() << 8);
	startAddress |= getc();

	length = (getc() << 8);
	length |= getc();

	for(i = 0; i < length; i++) {
		buffer[i] = getc();
	}

	// Download this code to the CPU.
	if(cpu_boot() != 0)
		return;
	
	destAddrHigh = startAddress >> 8;
	i = 0;

	while(i < length) {
		// LDA
		CLOCK6502 = 0;
		WAITQUARTER;
		//addr = PORTC & 0x3;
		//if(addr != 0) {printf("Unexpected addr a %d\n", i); RESET6502 = 0; return;}
		data_bus_write(0xa9); // LDA
		CLOCK6502 = 1;
		WAITQUARTER;

		// Argument to LDA
		CLOCK6502 = 0;
		WAITQUARTER;
		//addr = PORTC & 0x3;
		//if(addr != 1) {printf("Unexpected addr b %d\n", i); RESET6502 = 0; return;}
		data_bus_write(buffer[i]);
		CLOCK6502 = 1;
		WAITQUARTER;

		// Wait state for LDA
		// cpu_clock_full(1);

		// STA
		CLOCK6502 = 0;
		WAITQUARTER;
		//addr = PORTC & 0x3;
		//if(addr != 2) {printf("Unexpected addr c %d:%d\n", addr, i); RESET6502 = 0; return;}
		data_bus_write(0x8d); // STA
		CLOCK6502 = 1;
		WAITQUARTER;

		// Low byte for STA
		CLOCK6502 = 0;
		WAITQUARTER;
		//addr = PORTC & 0x3;
		//if(addr != 3) {printf("Unexpected addr d %d\n", i); RESET6502 = 0; return;}
		data_bus_write(i);
		CLOCK6502 = 1;
		WAITQUARTER;

		// High byte for STA
		CLOCK6502 = 0;
		WAITQUARTER;
		//addr = PORTC & 0x3;
		//if(addr != 0) {printf("Unexpected addr e %d\n", i); RESET6502 = 0; return;}
		data_bus_write(destAddrHigh);
		CLOCK6502 = 1;
		WAITQUARTER;

		// Wait states for STA
		cpu_clock_full(1);

		i += 1;
	}

	cpu_halt();

	printf("Uploaded %d bytes to %x: %x, %x, %x...\r\n", length, startAddress, buffer[0], buffer[1], buffer[2]);
}

void irq_handler(void) __interrupt 1 {
	// This gets the timer interrupt.
	
	// Enable the 6502 interrupt briefly.
	INT6502 = 0;
	WAITHALF;
	WAITHALF;
	INT6502 = 1;
	
	TMR1H = tmr1hval;
	TMR1L = tmr1lval;
	PIR1bits.TMR1IF = 0; // Timer interrupt flag clear
}

// Disable the ROM and run the CPU at full speed.
// This assumes something useful has been uploaded to the RAM.
void play()
{
	// Disconnect the ROM.
	ROMENABLE = 0;
	
	// Hold reset.
	RESET6502 = 0;

	// Clock high to start.
	CLOCK6502 = 0;

	// Wait two clocks for power to stabilise (if we just started)
	cpu_clock_full(2);

	// Bring reset up.
	RESET6502 = 1;

	cpu_clock_run(1);

	if(tmr1en) {
		// Enable the timer.
		T1CONbits.RD16 = tmr1rd16; // 16-bit reads
		TMR1H = tmr1hval; // 65536 - 20000 == overflow at 50MHz
		TMR1L = tmr1lval; // lower bit
		T1CONbits.T1CKPS1 = tmr1tickps1; // Prescale 10 = 1/4 of osc
		T1CONbits.T1CKPS0 = tmr1tickps0; // Lower bit of prescale
		T1CONbits.TMR1CS = tmr1cs; // Osc source: intosc / 4
		T1CONbits.T1OSCEN = 0; // External osc disabled
		
		// Interrupts.
		INTCONbits.PEIE = 1; // Peripheral interrupts on
		INTCONbits.GIE  = 1; // Global interrupt enable
		PIE1bits.TMR1IE = 1; // Timer interrupt enable
		PIR1bits.TMR1IF = 0; // Timer interrupt flag clear

		// drumroll...
		T1CONbits.TMR1ON = 1; // Timer enable
	}
}

// Stop the CPU and re-enable the ROM.
void stop()
{
	// Stop the timer.
	T1CONbits.TMR1ON = 0;

	cpu_clock_run(0);

	// Reconnect the ROM.
	ROMENABLE = 1;

	// Reset CPU
	RESET6502 = 0;
	cpu_clock_full(6);
}

void timer_test()
{
	unsigned int i;

	printf("Timer test\r\n");
	// Set the timer.
	T1CONbits.RD16 = 1; // 16-bit reads
	TMR1H = 0xb1; // 65536 - 20000 == overflow at 50MHz
	TMR1L = 0xe0; // lower bit
	T1CONbits.T1CKPS1 = 1; // Prescale 10 = 1/4 of osc
	T1CONbits.T1CKPS0 = 0; // Lower bit of prescale
	T1CONbits.TMR1CS = 0; // Osc source: intosc / 4
	T1CONbits.T1OSCEN = 0; // External osc disabled
	
	// Interrupts.
	INTCONbits.PEIE = 1; // Peripheral interrupts on
	INTCONbits.GIE  = 1; // Global interrupt enable
	PIE1bits.TMR1IE = 1; // Timer interrupt enable
	PIR1bits.TMR1IF = 0; // Timer interrupt flag clear

	// drumroll...
	T1CONbits.TMR1ON = 1; // Timer enable

	// Now wait a while
	for(i = 0; i < 65535; i++) 
		;

	T1CONbits.TMR1ON = 0;
}

void store_note()
{
	int i;
	for(i = 0; i < HASHLENGTH; i++) {
		hash[i] = getc();
	}
}

void display_note()
{
	int i;
	for(i = 0; i < HASHLENGTH; i++) {
		printf("%c", hash[i]);
	}
	printf("\r\n");
}

void reset_timer()
{
	tmr1hval = TMR1H_50HZ; // 65536 - 20000 = overflow at 50MHz
	tmr1lval = TMR1L_50HZ; // lower byte
	tmr1rd16 = 1; // 16-bit timer reads
	tmr1tickps1 = 1; // Prescale 10 = 1/4 of osc
	tmr1tickps0 = 0;
	tmr1cs = 0; // Osc source: intosc / 4
}

void set_timer()
{
	tmr1en = getc();

	if(tmr1en == 1) {
		tmr1hval = getc();
		tmr1lval = getc();
		tmr1rd16 = getc();
		tmr1tickps1 = getc();
		tmr1tickps0 = getc();
		tmr1cs = getc();
	}
}

void main()
{
	// 16MHz: set bits 6, 5, 4 of OSCCON.
	OSCCONbits.IRCF2 = 1;
	OSCCONbits.IRCF1 = 1;
	OSCCONbits.IRCF0 = 1;

	// Disable analog ports.
	ANSEL = 0x0;
	ANSELH = 0x0;

	// Port C: all output except for C0 and C1.
	TRISC = 0x03; 

	// 6502 /IRQ high -> no interrupt pending
	INT6502 = 1;

	// Port B: Pin 6 is output.
	TRISBbits.TRISB6 = 0;

	DATASTROBE = 0; // Strobe (for data bus buffer): low
	DATACLOCK = 0; // Clock (for data bus buffer): low

	ROMENABLE = 1;
	
	// Turn off PWM (why is it on?)
	CCP1CON = 0b00000000;

	// Magic numbers to set up serial, and then to set up a 9600 bps serial
	// line given a 4MHz system clock.
	/*
	TXSTAbits.TXEN = 1;
	TXSTAbits.SYNC = 0;
	TXSTAbits.BRGH = 0;
	RCSTAbits.SPEN = 1;
	BAUDCONbits.BRG16 = 0;
	SPBRGH = 0;
	SPBRG  = 5; // extra magical
	*/

	// Same idea but using the standard library and a 16MHz clock.
	// 115200... ish... baud is 8
	// 57600 ... ish... baud is 16
    usart_open(USART_TX_INT_OFF 
            & USART_RX_INT_OFF
            & USART_BRGH_HIGH
            & USART_ASYNCH_MODE
            & USART_EIGHT_BIT,
			16);
	stdout = STREAM_USART;

	// Enable receive circuitry.
	RCSTAbits.CREN = 1;

	// Default play speed.
	reset_timer();

	printf("Hello from PIC\r\n");
	while(1) {
		unsigned char cmd;

		/*
		int i;
		cmd = getc();
		data_bus_write(cmd);
		for(i = 0; i < 8; i++) {
			printf("%c", (cmd & 128) ? '1': '0');
			cmd <<= 1;
		}
		printf(" ");
		*/

		printf("\r\n>");

		cmd = getc();
		//printf("%c\r\n", cmd);

		switch(cmd) {
		case 't':
			cpu_test();
			break;
		case 'T':
			sid_test();
			break;
		case 'h':
			timer_test();
			break;
		case 'u':
			upload();
			break;
		case '5':
			/* Play at 50 Hz */
			reset_timer();
			tmr1hval = TMR1H_50HZ;
			tmr1lval = TMR1L_50HZ;
			play();
			break;
		case '6':
			/* Play at 60 Hz */
			reset_timer();
			tmr1hval = TMR1H_60HZ;
			tmr1lval = TMR1L_60HZ;
			play();
			break;
		case 'q':
			/* Custom timer values */
			set_timer();
			break;
		case 'p':
			play();
			break;
		case 's':
			stop();
			break;
		case 'n':
			store_note();
			break;
		case 'N':
			display_note();
			break;
		case '0':
		case '1':
		case '2':
			bus_test(cmd);
			break;
		default:
			printf("?\r\n");
			break;
		}

	}
}

